from luqum.elasticsearch import SchemaAnalyzer, ElasticsearchQueryBuilder
from luqum.parser import parser
from luqum.utils import UnknownOperationResolver


def simple_query_interpreter(request, queryset, view, q):
    # simple query
    filter_transform_method = f"filter_es_transform_query_{view.action.lower()}"
    if hasattr(view, filter_transform_method):
        q = getattr(view, filter_transform_method)(q)
    return queryset.query("multi_match", query=q, fields=["*"])


def luqum_query_interpreter(request, queryset, view, q):
    # luqum query
    return apply_luqum(queryset, q, view.document)


def apply_luqum(queryset, query, document):
    tree = parser.parse(query)
    resolver = UnknownOperationResolver()
    tree = resolver(tree)
    analyzer = SchemaAnalyzer(document._index.to_dict())
    message_es_builder = ElasticsearchQueryBuilder(**analyzer.query_builder_options())
    return queryset.query(message_es_builder(tree))



import contextlib
import contextlib
import logging
import threading

from django.conf import settings
from elasticsearch.helpers import bulk
from elasticsearch_dsl.connections import get_connection
from elasticsearch_dsl import Document

from .json import to_plain_json

log = logging.getLogger("django_es_drf")

es_lock = threading.local()


def es_locked():
    return getattr(es_lock, "locked", False)


def es_disabled():
    return getattr(es_lock, "disabled", False)


def es_index(obj, force=False):
    if es_disabled():
        return

    if es_locked():
        if not hasattr(es_lock, "objects"):
            es_lock.objects = []
        es_lock.objects.append((obj, False))
        es_flush(force)
    else:
        es_index_internal([(obj, False)])


def es_delete(obj, force=False):
    if es_disabled():
        return

    if es_locked():
        if not hasattr(es_lock, "objects"):
            es_lock.objects = []
        es_lock.objects.append((obj, True))
        es_flush(force)
    else:
        es_index_internal([(obj, True)])


def es_flush(force=False, refresh=False):
    objects = getattr(es_lock, "objects", [])
    if not force and len(objects) < getattr(settings, "ES_INDEX_BATCH", 200):
        return
    try:
        es_index_internal(objects, refresh=refresh)
    finally:
        es_lock.objects = []
        pass


def upsert(id, body, index):
    d = {}
    d["_op_type"] = "update"
    d["_index"] = index
    d["doc"] = body
    d["doc_as_upsert"] = True
    if id:
        d["_id"] = id
    return d


def es_index_internal(objects, refresh=True):
    from .document_registry import registry

    if not objects:
        return
    actions = []
    for obj, should_delete in objects:
        if isinstance(obj, Document):
            idx, _id, data = obj._index._name, obj.meta.id, to_plain_json(obj)
        else:
            try:
                idx, _id, data = registry.model_to_index_and_id_and_data(obj)
            except:
                log.exception("Error in indexing object %s:%s", type(obj), obj.pk)
                continue

        if should_delete and _id:
            actions.append({"_op_type": "delete", "_index": idx, "_id": _id})
        else:
            actions.append(upsert(id=_id, body=data, index=idx))

    ret = bulk(get_connection("default"), actions, stats_only=True, refresh=refresh)
    return ret


@contextlib.contextmanager
def bulk_es(bulk=True):
    es_lock.locked = bulk
    try:
        yield
        es_flush(True)
    finally:
        es_lock.locked = False


@contextlib.contextmanager
def disabled_es():
    es_lock.disabled = True
    try:
        yield
    finally:
        es_lock.disabled = False



from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from . import registry
from .indexer import es_delete, es_index


def data_saved(sender, instance=None, **kwargs):
    es_index(instance)


def data_deleted(sender, instance=None, **kwargs):
    es_delete(instance)


@receiver(m2m_changed)
def m2m_hook(sender, instance, action, **kwargs):
    if not action.startswith("post_"):
        return
    try:
        registry.get_registry_entry_from_django(type(instance))
        es_index(instance)
    except KeyError:
        return



from django.conf import settings
from rest_framework import fields, serializers
import elasticsearch_dsl as e

from django_es_drf.nested import object_builder, serializer_object_builder


def get(key, default):
    return getattr(settings, key, default)


DJANGO_ES_DEFAULT_FIELD_MAPPING = get(
    "DJANGO_ES_DEFAULT_FIELD_MAPPING",
    {
        fields.CharField: lambda fld_name, fld, ctx, **kwargs: e.Keyword(**kwargs),
        fields.IntegerField: lambda fld_name, fld, ctx, **kwargs: e.Integer(**kwargs),
        fields.FloatField: lambda fld_name, fld, ctx, **kwargs: e.Float(**kwargs),
        fields.DateTimeField: lambda fld_name, fld, ctx, **kwargs: e.Date(**kwargs),
        fields.DateField: lambda fld_name, fld, ctx, **kwargs: e.Date(**kwargs),
        fields.BooleanField: lambda fld_name, fld, ctx, **kwargs: e.Boolean(**kwargs),
        fields.NullBooleanField: lambda fld_name, fld, ctx, **kwargs: e.Boolean(
            **kwargs
        ),
        serializers.Serializer: object_builder,
        serializers.SerializerMethodField: lambda fld_name, fld, ctx, **kwargs: e.Keyword(
            **kwargs
        ),
    },
)


ES_DRF_DEFAULT_FIELD_MAPPING = get(
    "ES_DRF_DEFAULT_FIELD_MAPPING",
    {
        e.Keyword: lambda fld_name, fld, ctx, **kwargs: fields.CharField(
            **{"required": False, **kwargs}
        ),
        e.Text: lambda fld_name, fld, ctx, **kwargs: fields.CharField(
            **{"required": False, **kwargs}
        ),
        e.Integer: lambda fld_name, fld, ctx, **kwargs: fields.IntegerField(
            **{"required": False, **kwargs}
        ),
        e.Float: lambda fld_name, fld, ctx, **kwargs: fields.FloatField(
            **{"required": False, **kwargs}
        ),
        e.Date: lambda fld_name, fld, ctx, **kwargs: fields.DateTimeField(
            **{"required": False, **kwargs}
        ),
        e.Boolean: lambda fld_name, fld, ctx, **kwargs: fields.BooleanField(
            **{"required": False, **kwargs}
        ),
        e.Object: serializer_object_builder,
    },
)



from .document_registry import registry, DjangoDocument
from .drf.viewsets import ESViewSet
from .drf.aggs import AggBase, BucketAgg, TranslatedBucketAgg, NestedAgg, SeparatorAgg
from .drf.serializers import ESDocumentSerializer
from .drf.pagination import ESPagination, ESPage, ESPaginator
from .drf.backends.source import DynamicSourceBackend
from .drf.backends.filters import (
    ESAggsFilterBackend,
    QueryFilterBackend,
    BaseESFilterBackend,
)
from .drf.renderers import ESRenderer
from .indexer import bulk_es, disabled_es
from .json import ESEncoderClass, to_plain_json, es_dump


def g(obj, key):
    if key in obj:
        return obj[key]
    return None


def remove_nulls(d):
    for k, v in list(d.items()):
        if v is None:
            del d[k]
    return d
